@@ -3,6 +3,7 @@ package DateTime::Format::Natural;
 use strict;
 use warnings;
 use base qw(DateTime::Format::Natural::Base);
+use boolean qw(true false);
 
 use Carp qw(croak);
 use DateTime ();
@@ -10,7 +11,17 @@ use Date::Calc qw(Day_of_Week check_date);
 use List::MoreUtils qw(all any);
 use Params::Validate ':all';
 
-our $VERSION = '0.71';
+our $VERSION = '0.74';
+
+validation_options(
+    on_fail => sub
+{
+    my ($error) = @_;
+    chomp $error;
+    croak $error;
+},
+    stack_skip => 2,
+);
 
 sub new
 {
@@ -30,7 +41,7 @@ sub _init
 
     $self->{Format}        = $opts{format}        || 'd/m/y';
     $self->{Lang}          = $opts{lang}          || 'en';
-    $self->{Prefer_future} = $opts{prefer_future} || 0;
+    $self->{Prefer_future} = $opts{prefer_future} || false;
     $self->{Time_zone}     = $opts{time_zone}     || 'floating';
     $self->{Opts}{daytime} = $opts{daytime};
 
@@ -45,35 +56,25 @@ sub _init_check
 {
     my $self = shift;
 
-    validation_options(
-        on_fail => sub
-    {
-        my ($error) = @_;
-        chomp $error;
-        croak $error;
-    },
-        stack_skip => 2,
-    );
-
     validate(@_, {
         lang => {
             type => SCALAR,
-            optional => 1,
+            optional => true,
             regex => qr!^(?:en)$!,
         },
         format => {
             type => SCALAR,
-            optional => 1,
+            optional => true,
             regex => qr!^(?:[dmy]{1,4}[-./]){2}[dmy]{1,4}$!i,
         },
         prefer_future => {
-            type => SCALAR,
-            optional => 1,
-            regex => qr!^[01]$!,
+            # SCALARREF due to boolean.pm's implementation
+            type => BOOLEAN | SCALARREF,
+            optional => true,
         },
         time_zone => {
             type => SCALAR,
-            optional => 1,
+            optional => true,
             callbacks => {
                 'valid timezone' => sub
                 {
@@ -84,7 +85,7 @@ sub _init_check
         },
         daytime => {
             type => HASHREF,
-            optional => 1,
+            optional => true,
         },
     });
 }
@@ -112,7 +113,7 @@ sub parse_datetime
     if (scalar keys %count == 1 && $count{(keys %count)[0]} == 2) {
         if ($date_string =~ /^\S+\b\s+\b\S+/) {
             ($date_string, @{$self->{tokens}}) = split /\s+/, $date_string;
-            $self->{count}{tokens} = 1 + @{$self->{tokens}};
+            $self->{count}{tokens} = 1 + scalar @{$self->{tokens}};
         }
         else {
             $self->{count}{tokens} = 1;
@@ -164,14 +165,17 @@ sub parse_datetime
             return $self->_get_datetime_object;
         }
 
-        $self->{datetime}->set_year($year);
-        $self->{datetime}->set_month($month);
-        $self->{datetime}->set_day($day);
+        $self->{datetime}->set(
+            year  => $year,
+            month => $month,
+            day   => $day,
+        );
 
         $self->_set_valid_exp;
         $self->_set_modified(1);
 
         if (@{$self->{tokens} || []}) {
+            $self->{count}{tokens}--;
             $self->_unset_valid_exp;
             $self->_unset_modified;
             $self->_process;
@@ -193,13 +197,13 @@ sub _parse_init
     my $self = shift;
 
     if (@_ > 1) {
-        validate(@_, { string => 1 });
+        validate(@_, { string => { type => SCALAR }});
         my %opts             = @_;
         $self->{Date_string} = $opts{string};
         (undef)              = $opts{debug}; # legacy
     }
     else {
-        validate_pos(@_, 1);
+        validate_pos(@_, { type => SCALAR });
         ($self->{Date_string}) = @_;
     }
 
@@ -228,19 +232,19 @@ sub parse_datetime_duration
       ? split /\s+ $timespan_sep \s+/ix, $self->{Date_string}
       : ($self->{Date_string});
 
-    my @stack;
+    my @queue;
     foreach my $date_string (@date_strings) {
-        push @stack, $self->parse_datetime($date_string);
+        push @queue, $self->parse_datetime($date_string);
     }
 
-    return @stack;
+    return @queue;
 }
 
 sub success
 {
     my $self = shift;
 
-    return ($self->_get_valid_exp && !$self->_get_failure) ? 1 : 0;
+    return ($self->_get_valid_exp && !$self->_get_failure) ? true : false;
 }
 
 sub error
@@ -259,27 +263,29 @@ sub trace
 {
     my $self = shift;
 
-    my @modified;
-    foreach my $unit (grep { $_ ne 'total' } keys %{$self->{modified}}) {
-        push @modified, "$unit: $self->{modified}{$unit}";
-    }
-
-    return join "\n", @{$self->{trace}}, @modified;
+    return join "\n", @{$self->{trace}},
+      map  { my $unit = $_; "$unit: $self->{modified}{$unit}" }
+      grep { $_ ne 'total' } keys %{$self->{modified}};
 }
 
 sub _process
 {
     my $self = shift;
 
-    PARSE: foreach my $keyword (keys %{$self->{data}->__grammar('')}) {
+    if (!exists $self->{lookup}) {
+        foreach my $keyword (keys %{$self->{data}->__grammar('')}) {
+            push @{$self->{lookup}{scalar @{$self->{data}->__grammar($keyword)->[0]}}}, $keyword;
+        }
+    }
+
+    PARSE: foreach my $keyword (@{$self->{lookup}{$self->{count}{tokens}} || []}) {
+        last if $self->_get_modified >= $self->{count}{tokens};
+
         my @grammar = @{$self->{data}->__grammar($keyword)};
         my $types = shift @grammar;
 
-        next if @{$self->{tokens}} > @$types;
-        last if $self->_get_modified >= @{$self->{tokens}};
-
         foreach my $expression (@grammar) {
-            my $valid_expression = 1;
+            my $valid_expression = true;
             my $definition = $expression->[0];
             my @positions = keys %$definition;
             my %regex_stack;
@@ -290,7 +296,8 @@ sub _process
                             next;
                         }
                         else {
-                            $valid_expression = 0;
+                            $valid_expression = false;
+                            last;
                         }
                     }
                 }
@@ -301,7 +308,8 @@ sub _process
                         next;
                     }
                     else {
-                        $valid_expression = 0;
+                        $valid_expression = false;
+                        last;
                     }
                 }
                 else {
@@ -313,9 +321,9 @@ sub _process
                 $self->_set_valid_exp;
                 my $i;
                 foreach my $positions (@{$expression->[1]}) {
-                    my @values;
+                    my ($c, @values);
                     foreach my $pos (@$positions) {
-                        $values[$pos] = exists $regex_stack{$pos}
+                        $values[$c++] = exists $regex_stack{$pos}
                           ? $regex_stack{$pos}
                           : ${$self->_token($pos)};
                     }
@@ -336,7 +344,7 @@ sub _post_process_options
     my $self = shift;
 
     if ($self->{Prefer_future}) {
-        my %modified = map { $_ => 1 } grep { $_ ne 'total' } keys %{$self->{modified}};
+        my %modified = map { $_ => true } grep { $_ ne 'total' } keys %{$self->{modified}};
 
         if ($self->{count}{tokens} == 1
             && (any { $self->{tokens}->[0] =~ /$_/i } @{$self->{data}->{weekdays_all}})
@@ -352,8 +360,8 @@ sub _post_process_options
             && (exists $self->{modified}{month} && $self->{modified}{month} == 1)
             && (exists $self->{modified}{day}
                   ? $self->{modified}{day} == 1
-                    ? 1 : 0
-                  : 1)
+                    ? true : false
+                  : true)
             && ($self->{datetime}->day_of_year < DateTime->now->day_of_year)
         ) {
             $self->{postprocess}{year} = 1;
@@ -380,13 +388,13 @@ sub _get_error       { $_[0]->{error}         }
 sub _set_error       { $_[0]->{error} = $_[1] }
 sub _unset_error     { $_[0]->{error} = ''    }
 
-sub _get_failure     { $_[0]->{failure}     }
-sub _set_failure     { $_[0]->{failure} = 1 }
-sub _unset_failure   { $_[0]->{failure} = 0 }
+sub _get_failure     { $_[0]->{failure}         }
+sub _set_failure     { $_[0]->{failure} = true  }
+sub _unset_failure   { $_[0]->{failure} = false }
 
-sub _get_valid_exp   { $_[0]->{valid_expression}     }
-sub _set_valid_exp   { $_[0]->{valid_expression} = 1 }
-sub _unset_valid_exp { $_[0]->{valid_expression} = 0 }
+sub _get_valid_exp   { $_[0]->{valid_expression}         }
+sub _set_valid_exp   { $_[0]->{valid_expression} = true  }
+sub _unset_valid_exp { $_[0]->{valid_expression} = false }
 
 sub _get_modified    { $_[0]->{modified}{total} || 0     }
 sub _set_modified    { $_[0]->{modified}{total} += $_[1] }
@@ -437,7 +445,7 @@ sub _set_datetime
                                       hour      => $hour,
                                       minute    => $min,
                                       second    => $sec);
-    $self->{running_tests} = 1;
+    $self->{running_tests} = true;
 }
 
 1;
@@ -504,8 +512,8 @@ Specifies the format of numeric dates, defaults to 'C<d/m/y>'.
 
 =item * C<prefer_future>
 
-Turns ambigious weekdays/months to their futuristic relatives. Accepts a boolean,
-defaults to 0.
+Turns ambiguous weekdays/months to their futuristic relatives. Accepts a boolean,
+defaults to false.
 
 =item * C<time_zone>
 
@@ -514,7 +522,7 @@ recognized by L<DateTime>. Defaults to 'floating'.
 
 =item * C<daytime>
 
-A hash consisting of specific hours given for peculiar daytimes. Daytimes may be
+A hash reference consisting of customized daytime hours, which may be
 selectively changed.
 
 =back
@@ -554,16 +562,16 @@ string given.
 
 =head2 error
 
-Returns the error message if the parsing didn't succeed.
+Returns the error message if the parsing did not succeed.
 
 =head2 trace
 
-Returns a trace of methods which we're called within the Base class and
-a summary how often certain units were modified.
+Returns a trace of methods which were called within the Base class and
+a summary how often certain units have been modified.
 
 =head1 GRAMMAR
 
-The grammar handling has been rewritten to be easily extendable and hence 
+The grammar handling has been rewritten to be easily extendable and hence
 everybody is encouraged to propose sensible new additions and/or changes.
 
 See the classes C<DateTime::Format::Natural::Lang::[language_code]> if
@@ -571,7 +579,7 @@ you're intending to hack a bit on the grammar guts.
 
 =head1 EXAMPLES
 
-See the classes C<DateTime::Format::Natural::Lang::[language_code]> for a 
+See the classes C<DateTime::Format::Natural::Lang::[language_code]> for a
 overview of current valid input.
 
 =head1 CREDITS
@@ -597,6 +605,8 @@ valuable suggestions & patches:
  Jesse Vincent
  Jason May
  Pat Kale
+ Ankur Gupta
+ Alex Bowley
 
 =head1 SEE ALSO
 
